from insights.parsers import saphostexec, SkipException
from insights.parsers.saphostexec import SAPHostExecStatus, SAPHostExecVersion
from insights.tests import context_wrap
import pytest
import doctest

STATUS_DOC = """
saphostexec running (pid = 9159)
sapstartsrv running (pid = 9163)
saposcol running (pid = 9323)
""".strip()

STATUS_ABNORMAL = """
saphostexec running (pid = 9159)
sapstartsrv run (pid = 9163)
saposcol (pid = 9323)
""".strip()

VER_DOC = """
*************************** Component ********************
/usr/sap/hostctrl/exe/saphostexec: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
/usr/sap/hostctrl/exe/sapstartsrv: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
/usr/sap/hostctrl/exe/saphostctrl: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 04:43:56)
/usr/sap/hostctrl/exe/xml71d.so: 721, patch 1011, changelist 1814854, linuxx86_64, opt (Jan 13 2018, 01:12:10)
**********************************************************
--------------------
SAPHOSTAGENT information
--------------------
kernel release                721
kernel make variant           721_REL
compiled on                   Linux GNU SLES-9 x86_64 cc4.1.2  for linuxx86_64
compiled for                  64 BIT
compilation mode              Non-Unicode
compile time                  Jan 13 2018 04:40:52
patch number                  33
latest change number          1814854
---------------------
supported environment
---------------------
operating system
Linux 2.6
Linux 3
Linux
""".strip()

SHA_STOP = """
saphostexec stopped
""".strip()


def test_saphostexec_status_abnormal():
    with pytest.raises(SkipException) as s:
        SAPHostExecStatus(context_wrap(STATUS_ABNORMAL))
    assert "Incorrect status: 'sapstartsrv run (pid = 9163)'" in str(s)
    assert "Incorrect status: 'saposcol (pid = 9,23)'" not in str(s)


def test_saphostexec_status():
    sha_status = SAPHostExecStatus(context_wrap(STATUS_DOC))
    assert sha_status.is_running is True
    assert sha_status.services['saphostexec'] == '9159'
    assert 'saposcol' in sha_status

    sha_status = SAPHostExecStatus(context_wrap(SHA_STOP))
    assert sha_status.is_running is False
    assert sha_status.services == {}
    assert 'saposcol' not in sha_status


def test_saphostexec_version():
    sha_ver = SAPHostExecVersion(context_wrap(VER_DOC))
    assert sha_ver.components['saphostexec'].version == '721'
    assert sha_ver.components['saphostexec'].patch == '1011'
    assert sha_ver.components['xml71d.so'].changelist == '1814854'
    assert 'abc' not in sha_ver


def test_doc_examples():
    env = {
            'sha_status': SAPHostExecStatus(context_wrap(STATUS_DOC)),
            'sha_version': SAPHostExecVersion(context_wrap(VER_DOC)),
          }
    failed, total = doctest.testmod(saphostexec, globs=env)
    assert failed == 0



from __future__ import absolute_import
# !/usr/bin/python

from . import util
import json


class docker_wrapper:

    def __init__(self):
        cmd = ['docker', '-v']
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to communicate with the docker server')

    def inspect(self, obj_id):
        # returns dict representation of "docker inspect ID"
        cmd = ['docker', 'inspect', obj_id]
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to inspect object: %s' % obj_id)
        return json.loads(r.stdout)

    def driver(self):
        # returns the storage driver docker is using
        cmd = ['docker', 'info']
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to get docker info')

        for line in r.stdout.strip().split('\n'):
            if line.startswith('Storage Driver'):
                pre, _, post = line.partition(':')
                return post.strip()
        raise Exception('Unable to get docker storage driver')

    def dm_pool(self):
        # ONLY FOR DEVICEMAPPER
        # returns the docker-pool docker is using
        cmd = ['docker', 'info']
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to get docker info')
        for line in r.stdout.strip().split('\n'):
            if line.strip().startswith('Pool Name'):
                pre, _, post = line.partition(':')
                return post.strip()
        raise Exception('Unable to get docker pool name')

    def images(self, allI=False, quiet=False):
        # returns a list of dicts, each dict is an image's information
        # except when quiet is used - which returns a list of image ids
        # dict keys:
            # Created
            # Labels
            # VirtualSize
            # ParentId
            # RepoTags
            # RepoDigests
            # Id
            # Size
        cmd = ['docker', 'images', '-q', '--no-trunc']
        if allI:
            cmd.append("-a")
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to get docker images')
        images = r.stdout.strip().split('\n')
        if quiet:
            return images
        else:
            ims = []
            for i in images:
                inspec = self.inspect(i)
                inspec = inspec[0]
                dic = {}
                dic['Created'] = inspec['Created']
                if inspec['Config']:
                    dic['Labels'] = inspec['Config']['Labels']
                else:
                    dic['Labels'] = {}
                dic['VirtualSize'] = inspec['VirtualSize']
                dic['ParentId'] = inspec['Parent']
                dic['RepoTags'] = inspec['RepoTags']
                dic['RepoDigests'] = inspec['RepoDigests']
                dic['Id'] = inspec['Id']
                dic['Size'] = inspec['Size']
                ims.append(dic)
            return ims

    def containers(self, allc=False, quiet=False):
        # returns a list of dicts, each dict is an containers's information
        # except when quiet is used - which returns a list of container ids
        # dict keys:
            # Status
            # Created
            # Image
            # Labels
            # NetworkSettings
            # HostConfig
            # ImageID
            # Command
            # Names
            # Id
            # Ports

        cmd = ['docker', 'ps', '-q']
        if allc:
            cmd.append("-a")
        r = util.subp(cmd)
        if r.return_code != 0:
            raise Exception('Unable to get docker containers')
        containers = r.stdout.strip().split('\n')
        if quiet:
            return containers
        else:
            conts = []
            for i in containers:
                inspec = self.inspect(i)
                inspec = inspec[0]
                dic = {}
                dic['Status'] = inspec['State']['Status']
                dic['Created'] = inspec['Created']
                dic['Image'] = inspec['Config']['Image']
                dic['Labels'] = inspec['Config']['Labels']
                dic['NetworkSettings'] = inspec['NetworkSettings']
                dic['HostConfig'] = inspec['HostConfig']
                dic['ImageID'] = inspec['Image']
                dic['Command'] = inspec['Config']['Cmd']
                dic['Names'] = inspec['Name']
                dic['Id'] = inspec['Id']
                dic['Ports'] = inspec['NetworkSettings']['Ports']
                conts.append(dic)
            return conts



"""
saphostctrl - Commands ``saphostctrl``
======================================
"""
from insights import parser, CommandParser
from insights.parsers import ParseException, SkipException
from insights.specs import Specs
from insights.core.filters import add_filter


SAP_INST_FILTERS = [
        '******',
        'CreationClassName',
        'SID',
        'SystemNumber',
        'InstanceName',
        'Hostname',
        'SapVersionInfo',
]
add_filter(Specs.saphostctl_getcimobject_sapinstance, SAP_INST_FILTERS)


@parser(Specs.saphostctl_getcimobject_sapinstance)
class SAPHostCtrlInstances(CommandParser):
    """
    This class provides processing for the output of the
    ``/usr/sap/hostctrl/exe/saphostctrl -function GetCIMObject -enuminstances SAPInstance``
    command on SAP systems.

    Sample output of the command::

        *********************************************************
         CreationClassName , String , SAPInstance
         SID , String , D89
         SystemNumber , String , 88
         InstanceName , String , HDB88
         Hostname , String , hdb88
         FullQualifiedHostname , String , hdb88.example.com
         IPAddress , String , 10.0.0.88
         SapVersionInfo , String , 749, patch 211, changelist 1754007
        *********************************************************
         CreationClassName , String , SAPInstance
         SID , String , D90
         SystemNumber , String , 90
         InstanceName , String , HDB90
         Hostname , String , hdb90
         FullQualifiedHostname , String , hdb90.example.com
         IPAddress , String , 10.0.0.90
         SapVersionInfo , String , 749, patch 211, changelist 1754007

    Examples:
        >>> type(sap_inst)
        <class 'insights.parsers.saphostctrl.SAPHostCtrlInstances'>
        >>> sap_inst.data[-1]['CreationClassName']
        'SAPInstance'
        >>> sap_inst.data[-1]['SID']
        'D90'
        >>> sap_inst.data[-1]['SapVersionInfo']  # Note: captured as one string
        '749, patch 211, changelist 1754007'
        >>> sap_inst.data[0]['InstanceType']  # Inferred code from InstanceName
        'HDB'

    Attributes:
        data (list): List of dicts where keys are the lead name of each line and
            values are the string value.
        instances (list): The list of instances found in the cluster output.
        sids (list): The list of SID found in the cluster output.
        types (list): The list of instance types found in the cluster output.
    Raises:
        SkipException: When input is empty.
        ParseException: When input cannot be parsed.
    """
    def parse_content(self, content):
        if not content:
            raise SkipException("Empty content")
        if len(content) == 1:
            raise ParseException("Incorrect content: '{0}'".format(content))

        self.data = []
        self.instances = []
        _current_instance = {}
        _sids = set()
        _types = set()

        def _update_instance(inst):
            """
            This update and append is done twice - once when a new section
            is started and once at the end of the file.  So it's better to
            have it as a single function than to copy and paste the code.
            """
            _sids.add(inst['SID'])
            # InstanceType = The chars in InstanceName before the SystemNumber
            if ('InstanceName' in inst and 'SystemNumber' in inst and
                    'CreationClassName' in inst and 'Hostname' in inst and
                    'SapVersionInfo' in inst and 'SID' in inst and
                    inst['InstanceName'].endswith(inst['SystemNumber'])):
                # subtract len(sysnumber) characters from instance name
                itype = inst['InstanceName'][0:-len(inst['SystemNumber'])]
                inst['InstanceType'] = itype
                _types.add(itype)
            else:
                raise ParseException('Incorrect content.')
            # Now save the complete instance
            self.data.append(inst)
            self.instances.append(inst['InstanceName'])

        for line in (l.strip() for l in content):
            if line.startswith('******'):
                # Skip separator lines but save and reset current instance
                if _current_instance:
                    _update_instance(_current_instance)
                _current_instance = {}
                continue
            fields = line.split(' , ', 2)
            if len(fields) < 3:
                raise ParseException("Incorrect line: '{0}'".format(line))
            # TODO: if we see something other than 'String' in the second
            # field, we could maybe change its type, say to an integer?
            _current_instance[fields[0]] = fields[2]

        # There's no final line, so capture the current instance if we have one
        if _current_instance:
            _update_instance(_current_instance)
        self.sids = list(_sids)
        self.types = list(_types)

    def __len__(self):
        return len(self.data)



from insights.parsers import sysctl
from insights.tests import context_wrap
from insights.util import keys_in

SYSCTL_TEST = """
a=1
b = 2
c = include an = sign
""".strip()
SYSCTL_CONF_TEST = """
# sysctl.conf sample
#
  kernel.domainname = example.com

; this one has a space which will be written to the sysctl!
  kernel.modprobe = /sbin/mod probe
""".strip()


def test_sysctl():
    r = sysctl.Sysctl(context_wrap(SYSCTL_TEST))
    assert keys_in(["a", "b", "c"], r.data)
    assert r.data["a"] == "1"
    assert r.data["b"] == "2"
    assert r.data["c"] == "include an = sign"


def test_sysctl_conf():
    r = sysctl.SysctlConf(context_wrap(SYSCTL_CONF_TEST))
    assert keys_in(['kernel.domainname', 'kernel.modprobe'], r.data)
    assert r.data['kernel.domainname'] == 'example.com'
    assert r.data['kernel.modprobe'] == '/sbin/mod probe'



import pytest
from insights.parsers.date import Date, DateUTC, DateParseException
from insights.tests import context_wrap

DATE_OUTPUT1 = "Mon May 30 10:49:14 %s 2016"
DATE_OUTPUT2 = "Thu Oct 22 12:59:28 %s 2015"
DATE_OUTPUT_TRUNCATED = "Thu Oct 22"
DATE_OUTPUT_INVALID = "Fon Pla 34 27:63:89 CST 20-1"


def test_get_date1():
    DATE = DATE_OUTPUT1 % ('CST')
    date_info = Date(context_wrap(DATE))
    assert date_info.data == DATE
    assert date_info.datetime is not None
    assert date_info.timezone == 'CST'


def test_get_date2():
    DATE = DATE_OUTPUT2 % ('EDT')
    date_info = Date(context_wrap(DATE))
    assert date_info.data == DATE
    assert date_info.datetime is not None
    assert date_info.timezone == 'EDT'


def test_get_date_truncated():
    with pytest.raises(DateParseException) as e_info:
        Date(context_wrap(DATE_OUTPUT_TRUNCATED))
    assert 'Expected six date parts.  Got ' in str(e_info.value)


def test_get_date_invalid():
    with pytest.raises(DateParseException) as e_info:
        Date(context_wrap(DATE_OUTPUT_INVALID))
    assert DATE_OUTPUT_INVALID in str(e_info.value)


def test_get_date3():
    DATE = DATE_OUTPUT1 % ('UTC')
    date_info = DateUTC(context_wrap(DATE))
    assert date_info.data == DATE
    assert date_info.datetime is not None
    assert date_info.timezone == 'UTC'



"""
XFSInfo - command ``/usr/sbin/xfs_info {mount}``
================================================

The ``XFSInfo`` parser reads the output of the ``xfs_info`` command and turns
it into a dictionary of keys and values in several sections, as given in the
output of the command::

     meta-data=/dev/sda      isize=256    agcount=32, agsize=16777184 blks
              =              sectsz=512   attr=2
     data     =              bsize=4096   blocks=536869888, imaxpct=5
              =              sunit=32     swidth=128 blks
     naming   =version 2     bsize=4096
     log      =internal      bsize=4096   blocks=32768, version=2
              =              sectsz=512   sunit=32 blks, lazy-count=1
     realtime =none          extsz=524288 blocks=0, rtextents=0

The main sections are ``meta-data``, ``data``, ``naming``, ``log`` and
``realtime``, stored under those keys in the object's ``xfs_info`` property.
Each section can optionally have a '**specifier**', which is the first thing
after the section name (e.g. ``version`` or ``/dev/sda``).  If no specifier
is found on the line, none is recorded for the section.  The specifier can
also have a value (e.g. ``2`` for the version), which is recorded in the
``specifier_value`` key in the section.

Each 'key=value' pair until the next given section start (or end of file), is
recorded as an entry in the section dictionary, with all values that are
numeric being converted to integers (i.e. usually anything without a '`blks`'
suffix).

Because the spec for this parser can collect multiple files, the shared
parser information contains a list of XFSInfo objects, one per file system.

In addition, the ``data_size`` and ``log_size`` values are calculated as
properties from the block size and blocks in the data and log, respectively.

Attributes:
    xfs_info(dict): A dictionary of dictionaries containing the data from
        the report, keyed on the five section names in the output:
        '``meta-data``', '``data``', '``naming``', '``log``', and
        '``realtime``'.  '``meta-data``', '``data``' and '``log``' are
        always present.  Within each dictionary a special key
        '``specifier``' stores any data immediately after the section
        name - e.g. '/dev/sda' or 'version' in the case of the output
        below.  Any data immediately following that is stored in the
        ``specifier_value`` key.  Otherwise, data is read in key=value
        pairs - e.g. from the output below, the ``isize`` key will have
        the value ``32`` (an integer).  Data values given in blocks are
        left as is, so the value of the ``agsize`` key is '16777184 blks'
        as a string.
    mount(str): If the mount point can be derived from the file name of
        the original output, then this attribute contains the
        reconstructed mount point name.
    device(str): The device name immediately after the '``meta-data``'
        section heading.
    data_size(int): The size of the data segment in bytes, from
        multiplying the ``blocks`` and ``bsize`` values of the ``data``
        section together.
    log_size(int): The size of the log segment in bytes, from multiplying
        the ``blocks`` and ``bsize`` values of the ``log`` section
        together.

Sample output (from file '``sos_commands/xfs/xfs_info_.data``')::

     meta-data=/dev/sda      isize=256    agcount=32, agsize=16777184 blks
              =              sectsz=512   attr=2
     data     =              bsize=4096   blocks=536869888, imaxpct=5
              =              sunit=32     swidth=128 blks
     naming   =version 2     bsize=4096
     log      =internal      bsize=4096   blocks=32768, version=2
              =              sectsz=512   sunit=32 blks, lazy-count=1
     realtime =none          extsz=524288 blocks=0, rtextents=0

Examples:

    >>> xfs = shared[XFSInfo][0]  # first XFS filesystem as an example
    >>> xfs.xfs_info['meta-data']['specifier']
    '/dev/sda'
    >>> 'specifier_value' in xfs.xfs_info['meta-data']
    False
    >>> xfs.xfs_info['meta-data']['agcount']
    32
    >>> xfs.xfs_info['meta-data']['agsize']
    '16777184 blks'
    >>> xfs.data_size
    2199019061248
    >>> 'crc' in xfs.xfs_info['data']
    False

"""
from .. import parser, CommandParser
from insights.parsers import ParseException
import re
from insights.specs import Specs


@parser(Specs.xfs_info)
class XFSInfo(CommandParser):
    """
    This mapper reads the output of the ``xfs_info`` command.

    As this spec can collect more than one file, the mapper will return a
    list of XFSInfo objects, which need to be iterated through to find the
    information on the mount point or device you need.

    """

    def __init__(self, context):
        self.xfs_info = {}
        self.device = ''
        self.data_size = 0
        self.log_size = 0
        self.mount = ''

        # Determine mount from file name if possible:
        mount_name_re = re.compile('xfs_info_(?P<mount>.*)')
        match = mount_name_re.search(context.path)
        if match:
            # Try to put slashes back...
            self.mount = match.group('mount').replace('.', '/')

        super(XFSInfo, self).__init__(context)

    def parse_content(self, content):
        """
        In general the pattern is::

            section =sectionkey key1=value1 key2=value2, key3=value3
                    = key4=value4
            nextsec =sectionkey sectionvalue  key=value otherkey=othervalue

        Sections are continued over lines as per RFC822.  The first equals
        sign is column-aligned, and the first key=value is too, but the
        rest seems to be comma separated.  Specifiers come after the first
        equals sign, and sometimes have a value property, but sometimes not.
        """

        info_re = re.compile(r'^(?P<section>[\w-]+)?\s*' +
                             r'=(?:(?P<specifier>\S+)(?:\s(?P<specval>\w+))?)?' +
                             r'\s+(?P<keyvaldata>\w.*\w)$'
                             )
        keyval_re = re.compile(r'(?P<key>[\w-]+)=(?P<value>\d+(?: blks)?)')

        sect_info = None

        for line in content:
            match = info_re.search(line)
            if match:
                if match.group('section'):
                    # Change of section - make new sect_info dict and link
                    sect_info = {}
                    self.xfs_info[match.group('section')] = sect_info
                if match.group('specifier'):
                    sect_info['specifier'] = match.group('specifier')
                    if match.group('specval'):
                        sect_info['specifier_value'] = match.group('specval')
                for key, value in keyval_re.findall(match.group('keyvaldata')):
                    if value.isdigit():
                        # Convert numeric values to integers as a convenience
                        value = int(value)
                    sect_info[key] = value

        # Check that the meta-data and data sections are there and contain
        # the things we expect
        if 'meta-data' not in self.xfs_info:
            raise ParseException("No 'meta-data' section found")
        if 'specifier' not in self.xfs_info['meta-data']:
            raise ParseException("Device specifier not found in meta-data")
        if 'data' not in self.xfs_info:
            raise ParseException("No 'data' section found")
        if 'blocks' not in self.xfs_info['data']:
            raise ParseException("'blocks' not defined in data section")
        if 'bsize' not in self.xfs_info['data']:
            raise ParseException("'bsize' not defined in data section")
        if 'log' not in self.xfs_info:
            raise ParseException("No 'log' section found")
        if 'blocks' not in self.xfs_info['log']:
            raise ParseException("'blocks' not defined in log section")
        if 'bsize' not in self.xfs_info['log']:
            raise ParseException("'bsize' not defined in log section")

        # Store a few handy properties
        self.device = self.xfs_info['meta-data']['specifier']

        # Calculate a few things
        self.data_size = self.xfs_info['data']['blocks'] * self.xfs_info['data']['bsize']
        self.log_size = self.xfs_info['log']['blocks'] * self.xfs_info['log']['bsize']

    def __str__(self):
        return 'xfs_info of ' + self.xfs_info['meta-data']['specifier']



"""
Disk free space commands
========================

Module for the processing of output from the ``df`` command.  The base class
``DiskFree`` provides all of the functionality for all classes.
Data is avaliable as rows of the output contained
in one ``Record`` object for each line of output.

Sample input data for the ``df -li`` command looks like::

    Filesystem        Inodes  IUsed     IFree IUse% Mounted on
    /dev/mapper/vg_lxcrhel6sat56-lv_root
                     6275072 124955   6150117    2% /
    devtmpfs         1497120    532   1496588    1% /dev
    tmpfs            1499684    331   1499353    1% /dev/shm
    tmpfs            1499684    728   1498956    1% /run
    tmpfs            1499684     16   1499668    1% /sys/fs/cgroup
    tmpfs            1499684     54   1499630    1% /tmp
    /dev/sda2      106954752 298662 106656090    1% /home
    /dev/sda1         128016    429    127587    1% /boot
    tmpfs            1499684      6   1499678    1% /V M T o o l s
    tmpfs            1499684     15   1499669    1% /VM Tools

This module provides two parsers:

DiskFree_LI - command ``df -li``
--------------------------------

DiskFree_ALP - command ``df -alP``
----------------------------------

DiskFree_AL - command ``df -al``
--------------------------------

This example demonstrates the ``DiskFree_LI`` class but all classes will provide
the same functionality.

Examples:
    >>> df_info = shared[DiskFree_LI]
    >>> df_info.filesystem_names
    ['tmpfs', '/dev/mapper/vg_lxcrhel6sat56-lv_root', 'devtmpfs', '/dev/sda2', '/dev/sda1']
    >>> df_info.get_filesystem('/dev/sda2')
    [Record(filesystem='/dev/sda2', total='106954752', used='298662', available='106656090', capacity='1%', mounted_on='/home')]
    >>> df_info.mount_names
    ['/tmp', '/home', '/dev', '/boot', '/VM Tools', '/sys/fs/cgroup', '/', '/run', '/V M T o o l s', '/dev/shm']
    >>> df_info.get_mount('/boot')
    Record(filesystem='/dev/sda1', total='128016', used='429', available='127587', capacity='1%', mounted_on='/boot')
    >>> len(df_info)
    10
    >>> [d.mounted_on for d in df_info if 'sda' in d.filesystem]
    ['/home', '/boot']
    >>> df_info.data[0].filesystem
    '/dev/mapper/vg_lxcrhel6sat56-lv_root'
    >>> df_info.data[0]
    Record(filesystem='/dev/mapper/vg_lxcrhel6sat56-lv_root', total='6275072', used='124955', available='6150117', capacity='2%', mounted_on='/')
"""
from .. import parser, CommandParser
from collections import namedtuple, defaultdict

from insights.parsers import ParseException
from insights.specs import Specs

Record = namedtuple("Record", ['filesystem', 'total', 'used', 'available', 'capacity', 'mounted_on'])
"""namedtuple: Represents the information parsed from ``df`` command output."""


def parse_df_lines(df_content):
    """Parse contents of each line in ``df`` output.

    Parse each line of ``df`` output ensuring that wrapped lines are
    reassembled prior to parsing, and that mount names containing spaces
    are maintained.

    Parameters:
        df_content (list): Lines of df output to be parsed.

    Returns:
        list: A list of ``Record`` ``namedtuple``'s. One for each line of the
        ``df`` output with columns as the key values. The fields of
        ``Record`` provide information about the file system attributes
        as determined by the arguments to the ``df`` command.  So, for
        example, if ``df`` is given the ``-alP``, the values are in
        terms of 1024 blocks.  If ``-li`` is given, then the values are
        in terms of inodes::

            - filesystem: Name of the filesystem
            - total: total number of resources on the filesystem
            - used: number of the resources used on the filesystem
            - available: number of the resource available on the filesystem
            - capacity: percentage of the resource used on the filesystem
            - mounted_on: mount point of the filesystem
    """
    df_ls = {}
    df_out = []
    is_sep = False
    columns = Record._fields
    for line in df_content[1:]:  # [1:] -> Skip the header
        # Stop at 5 splits to avoid splitting spaces in path
        line_splits = line.rstrip().split(None, 5)
        if len(line_splits) >= 6:
            for i, name in enumerate(columns):
                df_ls[name] = line_splits[i]
            is_sep = False
        elif len(line_splits) == 1:
            # First line of the separated line
            df_ls[columns[0]] = line_splits[0]
            is_sep = True
        elif is_sep and len(line_splits) >= 5:
            # Re-split to avoid this kind of "Mounted on": "VMware Tools"
            line_splits = line.split(None, 4)
            # Last line of the separated line
            for i, name in enumerate(columns[1:]):
                df_ls[name] = line_splits[i]
            is_sep = False
        elif not line_splits:  # Skip empty lines (might in sosreport)
            continue
        else:
            raise ParseException("Could not parse line '{l}'".format(l=line))
        # Only add this record if we've got a line and it's not separated
        if df_ls and not is_sep:
            rec = Record(**df_ls)
            df_out.append(rec)
            df_ls = {}
    return df_out


class DiskFree(CommandParser):
    """Class to provide methods used by all ``df`` command classes.

    Attributes:
        data (list of Record): List of ``Record`` objects for each line of command
            output.
        filesystems (dict of list): Dictionary with each entry being a
            list of ``Record`` objects, for all lines in the command output. The
            dictionary is keyed by the ``filesystem`` value of the Record.
        mounts (dict): Dictionary with each entry being a ``Record`` object
            corresponding to the ``mounted_on`` key.
    """
    def __init__(self, context):
        super(DiskFree, self).__init__(context)
        filesystems = defaultdict(list)
        self.mounts = {}
        for datum in self.data:
            filesystems[datum.filesystem].append(datum)
            self.mounts[datum.mounted_on] = datum
        self.filesystems = dict(filesystems)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for row in self.data:
            yield row

    def parse_content(self, content):
        bad_lines = ["no such file or directory"]
        content = [l for l in content if bad_lines[0] not in l.lower()]
        self.data = parse_df_lines(content)

    @property
    def filesystem_names(self):
        """list: Returns list of unique filesystem names."""
        return self.filesystems.keys()

    def get_filesystem(self, name):
        """str: Returns list of Record objects for filesystem ``name``."""
        return self.filesystems.get(name, [])

    @property
    def mount_names(self):
        """list: Returns list of unique mount point names."""
        return self.mounts.keys()

    def get_mount(self, name):
        """Record: Returns Record obj for mount point ``name``."""
        return self.mounts.get(name)

    def get_dir(self, path):
        """
        Record: returns the Record object that contains the given path.

        This finds the most specific mount path that contains the given path.
        """
        try:
            longest = max(m for m in self.mounts if path.startswith(m))
            return self.mounts[longest]
        except ValueError:
            return None


@parser(Specs.df__li)
class DiskFree_LI(DiskFree):
    """Parse lines from the output of the ``df -li`` command.

    Typical content of the ``df -li`` command output looks like::

        Filesystem        Inodes  IUsed     IFree IUse% Mounted on
        /dev/mapper/vg_lxcrhel6sat56-lv_root
                         6275072 124955 6150117    2% /
        devtmpfs         1497120    532   1496588    1% /dev
        tmpfs            1499684    331   1499353    1% /dev/shm
        tmpfs            1499684    728   1498956    1% /tmp
        /dev/sda2      106954752 298662 106656090    1% /home
        /dev/sda1         128016    429    127587    1% /boot
        tmpfs            1499684      6   1499678    1% /run/user/988
        tmpfs            1499684     15   1499669    1% /run/user/100

    Attributes:
        data (list): A list of the ``df`` information with one ``Record`` object for
            each line of command output. Mapping of input columns to output
            fields is::

                Input column   Output Field
                ------------   ------------
                Filesystem     filesystem
                Inodes         total
                IUsed          used
                IFree          available
                IUse%          capacity
                Mounted on     mounted_on
    """
    pass


@parser(Specs.df__alP)
class DiskFree_ALP(DiskFree):
    """Parse lines from the output of the ``df -alP`` command.

    Typical content of the ``df -alP`` command looks like::

        Filesystem                           1024-blocks      Used Available Capacity Mounted on
        /dev/mapper/vg_lxcrhel6sat56-lv_root    98571884   4244032  89313940       5% /
        sysfs                                          0         0         0        - /sys
        proc                                           0         0         0        - /proc
        devtmpfs                                 5988480         0   5988480       0% /dev
        securityfs                                     0         0         0        - /sys/kernel/security
        tmpfs                                    5998736    491660   5507076       9% /dev/shm
        devpts                                         0         0         0        - /dev/pts
        tmpfs                                    5998736      1380   5997356       1% /run
        tmpfs                                    5998736         0   5998736       0% /sys/fs/cgroup
        cgroup                                         0         0         0        - /sys/fs/cgroup/systemd

    Attributes:
        data (list): A list of the ``df`` information with one ``Record`` object for
            each line of command output. Mapping of input columns to output
            fields is::

                Input column   Output Field
                ------------   ------------
                Filesystem     filesystem
                1024-blocks    total
                Used           used
                Available      available
                Capacity       capacity
                Mounted on     mounted_on
    """
    pass


@parser(Specs.df__al)
class DiskFree_AL(DiskFree):
    """Parse lines from the output of the ``df -al`` command.

    Typical content of the ``df -al`` command looks like::

        Filesystem                             1K-blocks      Used Available     Use% Mounted on
        /dev/mapper/vg_lxcrhel6sat56-lv_root    98571884   4244032  89313940       5% /
        sysfs                                          0         0         0        - /sys
        proc                                           0         0         0        - /proc
        devtmpfs                                 5988480         0   5988480       0% /dev
        securityfs                                     0         0         0        - /sys/kernel/security
        tmpfs                                    5998736    491660   5507076       9% /dev/shm
        devpts                                         0         0         0        - /dev/pts
        tmpfs                                    5998736      1380   5997356       1% /run
        tmpfs                                    5998736         0   5998736       0% /sys/fs/cgroup
        cgroup                                         0         0         0        - /sys/fs/cgroup/systemd

    Attributes:
        data (list): A list of the ``df`` information with one ``Record`` object for
            each line of command output. Mapping of input columns to output
            fields is::

                Input column   Output Field
                ------------   ------------
                Filesystem     filesystem
                1K-blocks      total
                Used           used
                Available      available
                Use%           capacity
                Mounted on     mounted_on
    """
    pass



from insights.tests import context_wrap
from insights.parsers.nmcli import NmcliDevShow
from insights.parsers.nmcli import NmcliConnShow
from insights.parsers import nmcli
import doctest

NMCLI_SHOW = """
GENERAL.DEVICE:                         em3
GENERAL.TYPE:                           ethernet
GENERAL.HWADDR:                         B8:2A:72:DE:F8:B9
GENERAL.MTU:                            1500
GENERAL.STATE:                          100 (connected)
GENERAL.CONNECTION:                     em3
GENERAL.CON-PATH:                       /org/freedesktop/NetworkManager/ActiveConnection/1
WIRED-PROPERTIES.CARRIER:               on
IP4.ADDRESS[1]:                         10.16.184.98/22
IP4.GATEWAY:                             10.16.187.254
IP4.DNS[1]:                             10.16.36.29
IP4.DNS[2]:                             10.11.5.19
IP4.DNS[3]:                             10.5.30.160
IP4.DOMAIN[1]:                          khw.lab.eng.bos.example.com
IP6.ADDRESS[1]:                         2620:52:0:10bb:ba2a:72ff:fede:f8b9/64
IP6.ADDRESS[2]:                         fe80::ba2a:72ff:fede:f8b9/64
IP6.GATEWAY:                            fe80:52:0:10bb::fc
IP6.ROUTE[1]:                           dst = 2620:52:0:10bb::/64, nh = ::, mt = 100

GENERAL.DEVICE:                         em1
GENERAL.TYPE:                           ethernet
GENERAL.HWADDR:                         B8:2A:72:DE:F8:BB
GENERAL.MTU:                            1500
GENERAL.STATE:                          100 connected
GENERAL.CONNECTION:                     --
GENERAL.CON-PATH:                       --
WIRED-PROPERTIES.CARRIER:               off

GENERAL.DEVICE:                         em2
GENERAL.TYPE:                           ethernet
GENERAL.HWADDR:                         B8:2A:72:DE:F8:BC
GENERAL.MTU:                            1500
GENERAL.STATE:                          100 (connected)
GENERAL.CONNECTION:                     --
GENERAL.CON-PATH:                       --
WIRED-PROPERTIES.CARRIER:               off
"""

NMCLI_SHOW_ERROR = """
Error: Option '-l' is unknown, try 'nmcli -help'.
"""

STATIC_CONNECTION_SHOW_1 = """
NAME      UUID                                  TYPE      DEVICE
enp0s3    320d4923-c410-4b22-b7e9-afc5f794eecc  ethernet  enp0s3
virbr0    7c7dec66-4a8c-4b49-834a-889194b3b83c  bridge    virbr0
test-net-1  f858b1cc-d149-4de0-93bc-b1826256847a  ethernet  --
""".strip()

STATIC_CONNECTION_SHOW_2 = """
NAME      UUID                                  TYPE      DEVICE
enp0s3    320d4923-c410-4b22-b7e9-afc5f794eecc  ethernet  enp0s3
virbr0    7c7dec66-4a8c-4b49-834a-889194b3b83c  bridge    virbr0
test-net-1  f858b1cc-d149-4de0-93bc-b1826256847a  ethernet  --
test-net-2 f858b1cc-d149-4de0-93bc-b1826256847a  ethernet  --
""".strip()


def test_nmcli():
    nmcli_obj = NmcliDevShow(context_wrap(NMCLI_SHOW))
    con_dev = nmcli_obj.connected_devices
    assert sorted(con_dev) == sorted(['em1', 'em3', 'em2'])
    assert nmcli_obj.data['em3']['IP4_GATEWAY'] == "10.16.187.254"
    assert nmcli_obj.data['em3']['IP4_DNS1'] == "10.16.36.29"
    assert nmcli_obj.data['em3']['IP6_ROUTE1'] == "dst = 2620:52:0:10bb::/64, nh = ::, mt = 100"
    assert nmcli_obj.data['em1']['STATE'] == "connected"
    assert nmcli_obj.data['em1']['CON-PATH'] == "--"
    assert nmcli_obj.data['em3']['IP6_ADDRESS1'] == "2620:52:0:10bb:ba2a:72ff:fede:f8b9/64"
    assert nmcli_obj.data['em3']['IP6_ADDRESS2'] == "fe80::ba2a:72ff:fede:f8b9/64"
    assert nmcli_obj.data['em3']['CON-PATH'] == "/org/freedesktop/NetworkManager/ActiveConnection/1"
    assert len(nmcli_obj.data['em3']) == 17
    assert len(nmcli_obj.data['em1']) == 7


def test_static_connection_test_1():
    static_conn = NmcliConnShow(context_wrap(STATIC_CONNECTION_SHOW_1))
    assert static_conn.data[0] == {'NAME': 'enp0s3', 'UUID': '320d4923-c410-4b22-b7e9-afc5f794eecc', 'TYPE': 'ethernet', 'DEVICE': 'enp0s3'}
    assert static_conn.disconnected_connection == ["test-net-1"]


def test_static_connection_test_2():
    static_conn = NmcliConnShow(context_wrap(STATIC_CONNECTION_SHOW_2))
    assert static_conn.disconnected_connection == ["test-net-1", "test-net-2"]


def test_nmcli_doc_examples():
    env = {
        'nmcli_obj': NmcliDevShow(context_wrap(NMCLI_SHOW)),
        'static_conn': NmcliConnShow(context_wrap(STATIC_CONNECTION_SHOW_1)),
    }
    failed, total = doctest.testmod(nmcli, globs=env)
    assert failed == 0



from __future__ import print_function
import argparse
import os
import random
import sys
from concurrent.futures import ThreadPoolExecutor

from insights.archive import repo, get_archives
from insights.util import fs

p = argparse.ArgumentParser("python -m insights.archive")
p.add_argument("which", choices=("demo", "integration"))
p.add_argument("-v", "--verbose", dest="verbose", action="store_true", default=False)
p.add_argument("-q", "--quiet", dest="quiet", action="store_true", default=False)
p.add_argument("-c", "--clean", dest="clean", action="store_true", default=False)
p.add_argument("-d", "--destination", dest="destination", default="")
p.add_argument("-u", "--upload", action="store_true", default=False)
p.add_argument("-H", "--host", default="localhost:8080")
p.add_argument("-a", "--auth", default="")
p.add_argument("-s", "--systems", default="")
p.add_argument("-x", "--exclude", type=int, default=0)
p.add_argument("package")

args = p.parse_args()
repo.console = not args.quiet
demo_anchovies = list(get_archives(args.package, args.systems))
random.shuffle(demo_anchovies)


def post_it(url, anchovy, file_name, args):
    r = session.post(url % anchovy.machine_id, files={"file": open(file_name, "rb")})
    if r.status_code == 201:
        print("*** Uploaded", anchovy.name)
        if args.verbose:
            print(json.dumps(r.json(), indent=4))
        else:
            print("Rule hits:", [str(rpt["rule_id"]) for rpt in r.json()["reports"]])
    else:
        print("*** Upload FAILED for %s: %d" % (anchovy.name, r.status_code))
        print(r.text)
        sys.exit(1)


for i in range(args.exclude):
    print("Excluding [%s]" % demo_anchovies.pop().name)

if args.which == "integration":
    dest = args.destination if args.destination else repo.DEFAULT_DEST
    if os.path.exists(dest) and args.clean:
        print("Cleaning integration tests.")
        fs.remove(dest)
    list(repo.build_integration_test_archives(args.package,
                                              machine_id="c21320b6-29b1-11e5-adc7-28d244603426"))
else:
    dest = args.destination if args.destination else "./demo-archives"
    if os.path.exists(dest) and args.clean:
        print("Cleaning demo archives.")
        fs.remove(dest)
    demo_anchovies.sort(key=lambda x: x.name)
    repo.build_all(demo_anchovies, dest=dest, clean=args.clean)
    if args.upload:
        import requests
        if args.verbose:
            import json
        session = requests.Session()
        session.headers = {"User-Agent": "Insights Demo Client"}
        session.auth = tuple(args.auth.split(":"))
        assert len(session.auth) == 2
        url = "http://%s" % args.host + "/r/insights/uploads/%s"
        pool = ThreadPoolExecutor(16)
        for anchovy in demo_anchovies:
            file_name = os.path.join(dest, anchovy.name + ".tar.gz")
            pool.submit(post_it, url, anchovy, file_name, args)
        pool.shutdown()
